"""Test helpers for pytest"""
from unittest.mock import patch

import pytest

from .const import TEST_V4_PATH


@pytest.fixture(autouse=True)
def mock_url():
    """Create 'mock_url' argument for test cases"""
    with patch("pytomorrowio.TomorrowioV4._get_url", return_value=TEST_V4_PATH):
        yield



"""Constants used by tests"""

TEST_V4_PATH = "/v4/timelines"

REALTIME_FIELDS_GREATER_THAN_MAX = [
    "temperature",
    "pressureSeaLevel",
    "humidity",
    "windSpeed",
    "windDirection",
    "pollutantO3",
    "weatherCode",
    "visibility",
    "windGust",
    "cloudCover",
    "precipitationType",
    "particulateMatter25",
    "particulateMatter10",
    "pollutantNO2",
    "pollutantCO",
    "pollutantSO2",
    "uvIndex",
    "uvHealthConcern",
    "epaIndex",
    "epaPrimaryPollutant",
    "epaHealthConcern",
    "mepIndex",
    "mepPrimaryPollutant",
    "mepHealthConcern",
    "treeIndex",
    "treeAcaciaIndex",
    "treeAshIndex",
    "treeBeechIndex",
    "treeBirchIndex",
    "treeCedarIndex",
    "treeCypressIndex",
    "treeElderIndex",
    "treeElmIndex",
    "treeHemlockIndex",
    "treeHickoryIndex",
    "treeJuniperIndex",
    "treeMahoganyIndex",
    "treeMapleIndex",
    "treeMulberryIndex",
    "treeOakIndex",
    "treePineIndex",
    "treeCottonwoodIndex",
    "treeSpruceIndex",
    "treeSycamoreIndex",
    "treeWalnutIndex",
    "treeWillowIndex",
    "weedIndex",
    "weedRagweedIndex",
    "grassIndex",
    "grassGrassIndex",
    "fireIndex",
    "temperatureApparent",
    "dewPoint",
    "pressureSurfaceLevel",
    "solarGHI",
    "cloudBase",
    "cloudCeiling",
]

CORE_FIELDS = [
    "temperature",
    "temperatureApparent",
    "dewPoint",
    "humidity",
    "windSpeed",
    "windGust",
    "pressureSurfaceLevel",
    "pressureSeaLevel",
    "precipitationIntensity",
    "rainIntensity",
    "freezingRainIntensity",
    "snowIntensity",
    "sleetIntensity",
    "precipitationProbability",
    "rainAccumulation",
    "snowAccumulation",
    "sleetAccumulation",
    "iceAccumulation",
    "visibility",
    "cloudCover",
    "cloudBase",
    "cloudCeiling",
]



"""Tomorrow.io field definitions"""
from dataclasses import dataclass
from datetime import timedelta
from typing import List

from .const import (
    ONE_DAY,
    ONE_HOUR,
    REALTIME,
    TYPE_AIR_QUALITY,
    TYPE_FIRE,
    TYPE_POLLEN,
    TYPE_PRECIPITATION,
    TYPE_SOLAR,
    TYPE_WEATHER,
)

MIN = "Min"
MAX = "Max"
AVG = "Avg"

ALL_MEASUREMENTS = [MIN, MAX, AVG]
NO_AVG = [MIN, MAX]


@dataclass
class FieldDefinition:
    """Tomorrow.io field definition"""

    max_timestep: timedelta
    measurements: List[str]
    type: str


FIELDS_V4 = {
    "temperature": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_WEATHER,
    ),
    "temperatureApparent": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_WEATHER,
    ),
    "dewPoint": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_WEATHER,
    ),
    "humidity": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_WEATHER,
    ),
    "windSpeed": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_WEATHER,
    ),
    "windDirection": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=[AVG],
        type=TYPE_WEATHER,
    ),
    "windGust": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_WEATHER,
    ),
    "pressureSurfaceLevel": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_WEATHER,
    ),
    "pressureSeaLevel": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_WEATHER,
    ),
    "precipitationIntensity": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_PRECIPITATION,
    ),
    "precipitationProbability": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_PRECIPITATION,
    ),
    "precipitationType": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=[],
        type=TYPE_PRECIPITATION,
    ),
    "hailBinary": FieldDefinition(
        max_timestep=REALTIME,
        measurements=[],
        type=TYPE_PRECIPITATION,
    ),
    "solarGHI": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_SOLAR,
    ),
    "solarDNI": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_SOLAR,
    ),
    "solarDHI": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_SOLAR,
    ),
    "visibility": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_WEATHER,
    ),
    "cloudCover": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_WEATHER,
    ),
    "cloudBase": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_WEATHER,
    ),
    "cloudCeiling": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_WEATHER,
    ),
    "weatherCode": FieldDefinition(
        max_timestep=ONE_DAY,
        measurements=NO_AVG,
        type=TYPE_WEATHER,
    ),
    "particulateMatter25": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_AIR_QUALITY,
    ),
    "particulateMatter10": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_AIR_QUALITY,
    ),
    "pollutantO3": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_AIR_QUALITY,
    ),
    "pollutantNO2": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_AIR_QUALITY,
    ),
    "pollutantCO": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_AIR_QUALITY,
    ),
    "pollutantSO2": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_AIR_QUALITY,
    ),
    "mepIndex": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_AIR_QUALITY,
    ),
    "mepPrimaryPollutant": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=[],
        type=TYPE_AIR_QUALITY,
    ),
    "mepHealthConcern": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_AIR_QUALITY,
    ),
    "epaIndex": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_AIR_QUALITY,
    ),
    "epaPrimaryPollutant": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=[],
        type=TYPE_AIR_QUALITY,
    ),
    "epaHealthConcern": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_AIR_QUALITY,
    ),
    "treeIndex": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_POLLEN,
    ),
    "grassIndex": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_POLLEN,
    ),
    "grassGrassIndex": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_POLLEN,
    ),
    "weedIndex": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_POLLEN,
    ),
    "weedRagweedIndex": FieldDefinition(
        max_timestep=ONE_HOUR,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_POLLEN,
    ),
    "fireIndex": FieldDefinition(
        max_timestep=REALTIME,
        measurements=ALL_MEASUREMENTS,
        type=TYPE_FIRE,
    ),
}



"""Tests for `pytomorrowio` package."""
import re
from datetime import datetime
from http import HTTPStatus
from typing import Mapping, Sequence
from unittest.mock import patch

import pytest
from aiohttp import ClientConnectionError, ClientResponseError, ClientSession

from pytomorrowio import TomorrowioV4
from pytomorrowio.const import (
    FIVE_MINUTES,
    MAX_FIELDS_PER_REQUEST,
    ONE_DAY,
    ONE_HOUR,
    ONE_MINUTE,
    REALTIME,
    TYPE_POLLEN,
    TYPE_PRECIPITATION,
    TYPE_WEATHER,
)
from pytomorrowio.exceptions import (
    CantConnectException,
    InvalidAPIKeyException,
    InvalidTimestep,
    MalformedRequestException,
    RateLimitedException,
    UnknownException,
)

from .const import CORE_FIELDS, REALTIME_FIELDS_GREATER_THAN_MAX
from .helpers import create_session, create_trace_config

GPS_COORD = (28.4195, -81.5812)


async def _test_capture_request_and_response():
    # Remove leading underscore to capture request & response to stdout
    async with ClientSession(trace_configs=[create_trace_config()]) as session:
        api = TomorrowioV4("real_api_key", *GPS_COORD, session=session)

        await api.realtime_and_all_forecasts(
            realtime_fields=api.available_fields(REALTIME),
            nowcast_fields=api.available_fields(ONE_MINUTE),
            hourly_fields=api.available_fields(ONE_HOUR),
            daily_fields=api.available_fields(ONE_DAY),
            nowcast_timestep=1,
        )

        assert False  # force traces to be displayed


async def test_raises_malformed_request(aiohttp_client):
    session = await create_session(
        aiohttp_client, "timelines_1hour.json", status=HTTPStatus.BAD_REQUEST
    )

    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)
    available_fields = api.available_fields(
        ONE_HOUR, [TYPE_POLLEN, TYPE_PRECIPITATION, TYPE_WEATHER]
    )

    assert api.max_requests_per_day == 100
    assert api.num_api_requests == 0

    with pytest.raises(MalformedRequestException):
        await api.forecast_hourly(available_fields)


async def test_raises_invalid_api_key(aiohttp_client):
    session = await create_session(
        aiohttp_client, "timelines_1hour.json", status=HTTPStatus.UNAUTHORIZED
    )

    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)
    available_fields = api.available_fields(
        ONE_HOUR, [TYPE_POLLEN, TYPE_PRECIPITATION, TYPE_WEATHER]
    )

    assert api.max_requests_per_day == 100
    assert api.num_api_requests == 0

    with pytest.raises(InvalidAPIKeyException):
        await api.forecast_hourly(available_fields)


async def test_raises_rate_limited(aiohttp_client):
    headers = {
        "RateLimit-Limit": "3",
        "RateLimit-Remaining": "0",
        "RateLimit-Reset": "1",
        "X-RateLimit-Limit-Day": "500",
        "X-RateLimit-Limit-Hour": "25",
        "X-RateLimit-Limit-Second": "3",
        "X-RateLimit-Remaining-Day": "484",
        "X-RateLimit-Remaining-Hour": "22",
        "X-RateLimit-Remaining-Second": "0",
    }

    session = await create_session(
        aiohttp_client,
        "timelines_1hour.json",
        headers=headers,
        status=HTTPStatus.TOO_MANY_REQUESTS,
    )

    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)
    available_fields = api.available_fields(
        ONE_HOUR, [TYPE_POLLEN, TYPE_PRECIPITATION, TYPE_WEATHER]
    )

    assert api.max_requests_per_day == 100
    assert api.num_api_requests == 0

    with pytest.raises(RateLimitedException):
        await api.forecast_hourly(available_fields)

    assert api.rate_limits.get("RateLimit-Reset") == 1
    assert api.rate_limits.get("X-RateLimit-Remaining-Second") == 0
    assert api.rate_limits.get("X-RateLimit-Remaining-Hour") == 22
    assert api.rate_limits.get("X-RateLimit-Remaining-Day") == 484
    assert api.max_requests_per_day == 500


async def test_timelines_hourly_good(aiohttp_client):
    session = await create_session(aiohttp_client, "timelines_1hour.json")

    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)
    available_fields = api.available_fields(
        ONE_HOUR, [TYPE_POLLEN, TYPE_PRECIPITATION, TYPE_WEATHER]
    )

    res = await api.forecast_hourly(available_fields)

    assert api.num_api_requests == 1

    assert res is not None
    assert isinstance(res, list)
    assert len(res) == 109

    for interval in res:
        assert isinstance(interval, Mapping)

        start_time = interval.get("startTime")
        assert isinstance(start_time, str)

        # Might be fixed in Python 3.11 - https://github.com/python/cpython/issues/80010
        # Maybe pytomorrowio should do the replacements to improve user experience
        start_time = re.sub("Z$", "+00:00", start_time)
        datetime.fromisoformat(start_time)

        values = interval.get("values")
        assert isinstance(values, Mapping)

        assert set(values) == set(available_fields)


async def test_timelines_daily_good(aiohttp_client):
    session = await create_session(aiohttp_client, "timelines_1day.json")

    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)
    available_fields = api.convert_fields_to_measurements(
        api.available_fields(ONE_DAY, [TYPE_POLLEN, TYPE_PRECIPITATION, TYPE_WEATHER])
    )
    res = await api.forecast_daily(available_fields)

    assert api.num_api_requests == 1

    assert res is not None
    assert isinstance(res, list)

    assert len(res) == 15

    for interval in res:
        assert isinstance(interval, Mapping)

        start_time = interval.get("startTime")
        assert isinstance(start_time, str)

        values = interval.get("values")
        assert isinstance(values, Mapping)

        assert set(values) == set(available_fields)


async def test_timelines_daily_greater_than_max_fields_good(aiohttp_client):
    fields = []
    for field in CORE_FIELDS:
        fields.append(f"{field}Min")
        fields.append(f"{field}Max")
        fields.append(field)
    session = await create_session(
        aiohttp_client,
        [
            "timelines_1day_greater_than_max_fields_1.json",
            "timelines_1day_greater_than_max_fields_2.json",
        ],
    )

    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)
    res = await api.forecast_daily(fields)

    assert api.num_api_requests == 2

    assert res is not None
    assert isinstance(res, list)

    assert len(res) == 16

    for interval in res:
        assert isinstance(interval, Mapping)

        start_time = interval.get("startTime")
        assert isinstance(start_time, str)

        values = interval.get("values")
        assert isinstance(values, Mapping)

        assert set(values) == set(fields)


async def test_timelines_5min_good(aiohttp_client):
    session = await create_session(aiohttp_client, "timelines_5min.json")

    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)
    available_fields = api.available_fields(
        FIVE_MINUTES, [TYPE_POLLEN, TYPE_PRECIPITATION, TYPE_WEATHER]
    )
    res = await api.forecast_nowcast(available_fields)

    assert api.num_api_requests == 1

    assert res is not None
    assert isinstance(res, list)

    assert len(res) == 73

    for interval in res:
        assert isinstance(interval, Mapping)

        start_time = interval.get("startTime")
        assert isinstance(start_time, str)

        values = interval.get("values")
        assert isinstance(values, Mapping)

        assert set(values) == set(available_fields)


async def test_timelines_realtime_good(aiohttp_client):
    session = await create_session(aiohttp_client, "timelines_realtime.json")

    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)
    available_fields = api.available_fields(FIVE_MINUTES)
    res = await api.realtime(available_fields)

    assert api.num_api_requests == 1

    assert res is not None
    assert isinstance(res, Mapping)


async def test_timelines_realtime_greater_than_max_fields_good(aiohttp_client):
    session = await create_session(
        aiohttp_client,
        [
            "timelines_realtime_more_than_max_fields_1.json",
            "timelines_realtime_more_than_max_fields_2.json",
        ],
    )
    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)

    res = await api.realtime(REALTIME_FIELDS_GREATER_THAN_MAX)

    assert api.num_api_requests == 2

    assert res is not None
    assert isinstance(res, Mapping)

    assert set(res) == set(REALTIME_FIELDS_GREATER_THAN_MAX)
    assert len(res) > MAX_FIELDS_PER_REQUEST


async def test_timelines_realtime_and_nowcast_good(aiohttp_client):
    session = await create_session(
        aiohttp_client,
        ["timelines_realtime_1min_1hour_1day.json", "timelines_realtime.json"],
    )

    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)

    res = await api.realtime_and_all_forecasts(
        realtime_fields=api.available_fields(REALTIME),
        all_forecasts_fields=api.available_fields(ONE_MINUTE),
        nowcast_timestep=1,
    )

    assert api.num_api_requests == 2

    assert res is not None
    assert isinstance(res, Mapping)

    current = res.get("current")
    assert isinstance(current, Mapping)
    assert current.get("temperature") == 74.53

    forecasts = res.get("forecasts")
    assert isinstance(forecasts, Mapping)
    assert set(forecasts.keys()) == {"hourly", "nowcast", "daily"}

    for key, expected_count in {"nowcast": 721, "hourly": 360, "daily": 16}.items():
        forecast = forecasts[key]
        assert isinstance(forecast, Sequence)
        assert len(forecast) == expected_count


async def test_timelines_realtime_nowcast_hourly_daily(aiohttp_client):
    session = await create_session(
        aiohttp_client,
        [
            "timelines_1min.json",
            "timelines_1hour.json",
            "timelines_1day.json",
            "timelines_realtime.json",
        ],
    )

    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)

    res = await api.realtime_and_all_forecasts(
        realtime_fields=api.available_fields(REALTIME),
        nowcast_fields=api.available_fields(ONE_MINUTE),
        hourly_fields=api.available_fields(ONE_HOUR),
        daily_fields=api.available_fields(ONE_DAY),
        nowcast_timestep=1,
    )

    assert api.num_api_requests == 4

    assert res is not None
    assert isinstance(res, Mapping)

    current = res.get("current")
    assert isinstance(current, Mapping)
    assert current.get("temperature") == 74.53

    forecasts = res.get("forecasts")
    assert isinstance(forecasts, Mapping)
    assert set(forecasts.keys()) == {"hourly", "nowcast", "daily"}

    for key, expected_count in {"nowcast": 361, "hourly": 109, "daily": 15}.items():
        forecast = forecasts[key]
        assert isinstance(forecast, Sequence)
        print(key, len(forecast))
        assert len(forecast) == expected_count


async def test_errors(aiohttp_client):
    """Test errors."""
    with pytest.raises(ValueError, match="unit_system"):
        TomorrowioV4("bogus_api_key", 0, 0, "fake_unit")

    with pytest.raises(InvalidTimestep):
        TomorrowioV4.available_fields("not_a_timestep", [])

    session = await create_session(
        aiohttp_client, "empty_response.json", status=HTTPStatus.BAD_GATEWAY
    )
    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)
    with pytest.raises(ClientResponseError):
        await api.realtime(["test"])

    session = await create_session(
        aiohttp_client, "empty_response.json", status=HTTPStatus.PERMANENT_REDIRECT
    )
    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)
    with pytest.raises(UnknownException):
        await api.realtime(["test"])

    with pytest.raises(InvalidTimestep):
        await api.forecast_nowcast(["test"], timestep=99)

    with pytest.raises(ValueError, match="Either"):
        await api.realtime_and_all_forecasts(
            ["test"], all_forecasts_fields=["test"], nowcast_fields=["test"]
        )

    with pytest.raises(ValueError, match="At least"):
        await api.realtime_and_all_forecasts(["test"])

    with patch(
        "pytomorrowio.pytomorrowio.ClientSession.post",
        side_effect=ClientConnectionError,
    ), pytest.raises(CantConnectException):
        await api.realtime(["test"])

    session = await create_session(aiohttp_client, "empty_response.json")
    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)
    with pytest.raises(UnknownException):
        await api.realtime(["test"])

    session = await create_session(aiohttp_client, "empty_response.json")
    api = TomorrowioV4("bogus_api_key", *GPS_COORD, session=session)
    with pytest.raises(UnknownException):
        await api.forecast_nowcast(["test"])



"""Main module."""
import asyncio
import json
import logging
from datetime import datetime, timedelta, timezone
from http import HTTPStatus
from typing import Any, Dict, List, Optional, Union

from aiohttp import ClientConnectionError, ClientSession
from multidict import CIMultiDict, CIMultiDictProxy

from .const import (
    BASE_URL_V4,
    CURRENT,
    DAILY,
    FIFTEEN_MINUTES,
    FIVE_MINUTES,
    FORECASTS,
    HEADER_DAILY_API_LIMIT,
    HEADER_REMAINING_CALLS_IN_SECOND,
    HEADERS,
    HOURLY,
    MAX_FIELDS_PER_REQUEST,
    NOWCAST,
    ONE_DAY,
    ONE_HOUR,
    ONE_MINUTE,
    THIRTY_MINUTES,
    TIMESTEP_CURRENT,
    TIMESTEP_DAILY,
    TIMESTEP_HOURLY,
    VALID_TIMESTEPS,
)
from .exceptions import (
    CantConnectException,
    InvalidAPIKeyException,
    InvalidTimestep,
    MalformedRequestException,
    RateLimitedException,
    UnknownException,
)
from .fields import FIELDS_V4
from .helpers import async_to_sync

_LOGGER = logging.getLogger(__name__)


def process_v4_fields(
    fields: List[str], timestep: timedelta, write_log: bool = True
) -> List[str]:
    """
    Filter v4 field list to only include valid fields for a given endpoint.

    Optionally logs a warning when fields get filtered out.
    """
    valid_fields = [field for field in fields if field in FIELDS_V4]
    if write_log and len(valid_fields) < len(fields):
        _LOGGER.warning(
            "Removed invalid fields: %s", list(set(fields) - set(valid_fields))
        )
    processed_fields = [
        field for field in valid_fields if timestep <= FIELDS_V4[field].max_timestep
    ]
    if write_log and len(processed_fields) < len(valid_fields):
        _LOGGER.warning(
            "Removed fields not available for `%s` timestep: %s",
            timestep,
            list(set(valid_fields) - set(processed_fields)),
        )
    return processed_fields


def dt_to_utc(input_dt: datetime) -> datetime:
    """If input datetime has a timezone defined, convert to UTC."""
    if input_dt and input_dt.tzinfo:
        return input_dt.astimezone(timezone.utc)
    return input_dt


def _timedelta_to_str(timestep: timedelta) -> str:
    """Convert timedelta to timestep string."""
    if timestep == ONE_DAY:
        return TIMESTEP_DAILY
    if timestep == ONE_HOUR:
        return TIMESTEP_HOURLY
    if timestep not in (THIRTY_MINUTES, FIFTEEN_MINUTES, FIVE_MINUTES, ONE_MINUTE):
        raise InvalidTimestep(f"Invalid `timestep` value {timestep}")
    return f"{int(timestep.total_seconds()/60)}m"


def _timestep_to_key(timestep: str) -> str:
    """Convert timestep to dict key."""
    if timestep == TIMESTEP_DAILY:
        return DAILY
    if timestep == TIMESTEP_HOURLY:
        return HOURLY
    return NOWCAST


class TomorrowioV4:
    """Async class to query the Tomorrow.io v4 API."""

    def __init__(
        self,
        apikey: str,
        latitude: Union[int, float, str],
        longitude: Union[int, float, str],
        unit_system: str = "imperial",
        session: Optional[ClientSession] = None,
    ) -> None:
        """Initialize Tomorrow.io API object."""
        if unit_system.lower() not in ("metric", "imperial"):
            raise ValueError("`unit_system` must be `metric` or `imperial`")

        self.api_key = apikey
        self.unit_system = unit_system.lower()
        self._session = session
        self._params = {
            "location": f"{float(latitude)},{float(longitude)}",
            "units": self.unit_system,
        }
        self._headers = {**HEADERS, "apikey": self.api_key}
        self._rate_limits: CIMultiDict = CIMultiDict()
        self._num_api_requests: int = 0

    @property
    def rate_limits(self) -> CIMultiDictProxy:
        """Return tomorrow.io rate limits for API key"""
        return CIMultiDictProxy(self._rate_limits)  # make read-only

    @property
    def max_requests_per_day(self) -> int:
        """
        Return the maximum number of requests per day.

        Defaults to 100 as a safe fallback.
        """
        return self.rate_limits.get(HEADER_DAILY_API_LIMIT, 100)

    @property
    def _remaining_requests_in_second(self) -> int:
        """
        Return the max remaining requests that can be made in the current seconed.

        Defaults to 1 so the first call can be made.
        """
        return self.rate_limits.get(HEADER_REMAINING_CALLS_IN_SECOND, 1)

    @property
    def num_api_requests(self) -> int:
        """The number of API requests made during the most recent call."""
        return self._num_api_requests

    @staticmethod
    def convert_fields_to_measurements(fields: List[str]) -> List[str]:
        """Converts general field list into fields with measurements."""
        field_list = []
        for field in fields:
            measurements = FIELDS_V4[field].measurements
            if len(measurements) < 2:
                field_list.append(field)
            else:
                field_list.extend(
                    [f"{field}{measurement}" for measurement in measurements]
                )

        return field_list

    @staticmethod
    def available_fields(
        timestep: timedelta, types: Optional[List[str]] = None
    ) -> List[str]:
        "Return available fields for a given timestep."
        if timestep not in VALID_TIMESTEPS:
            raise InvalidTimestep(f"{timestep} is not a valid 'timestep' parameter")
        fields = process_v4_fields(list(FIELDS_V4), timestep, write_log=False)

        if types:
            return [field for field in fields if FIELDS_V4[field].type in types]

        return fields

    @staticmethod
    def _get_url() -> str:
        # This method is required for test mocks
        return BASE_URL_V4

    async def _call_api(self, params: Dict[str, Any]) -> Dict[str, Any]:
        if self._session:
            return await self._make_call(params, self._session)

        async with ClientSession() as session:
            return await self._make_call(params, session)

    async def _make_call(
        self, params: Dict[str, Any], session: ClientSession
    ) -> Dict[str, Any]:
        if self._remaining_requests_in_second == 0:
            await asyncio.sleep(1)

        try:
            resp = await session.post(
                self._get_url(),
                headers=self._headers,
                data=json.dumps({**self._params, **params}),
            )
            resp_json = await resp.json()
        except ClientConnectionError as error:
            raise CantConnectException() from error

        self._rate_limits = CIMultiDict(
            {k: int(v) for k, v in resp.headers.items() if "ratelimit" in k.lower()}
        )

        if resp.status in (HTTPStatus.OK, HTTPStatus.PARTIAL_CONTENT):
            self._num_api_requests += 1
            for warning in set(
                warning["message"] for warning in resp_json.get("warnings", [])
            ):
                _LOGGER.info(
                    (
                        "While calling the API for the timesteps [%s], the following "
                        "warning was returned: %s"
                    ),
                    ", ".join(params["timesteps"]),
                    warning,
                )

            return resp_json
        if resp.status == HTTPStatus.BAD_REQUEST:
            raise MalformedRequestException(resp_json, resp.headers)
        if resp.status in (HTTPStatus.UNAUTHORIZED, HTTPStatus.FORBIDDEN):
            raise InvalidAPIKeyException(resp_json, resp.headers)
        if resp.status == HTTPStatus.TOO_MANY_REQUESTS:
            raise RateLimitedException(resp_json, resp.headers)

        resp.raise_for_status()
        raise UnknownException(resp_json, resp.headers)

    async def realtime(
        self,
        fields: List[str],
        reset_num_api_requests: bool = True,
        **additional_params,
    ) -> Dict[str, Any]:
        """Return realtime weather conditions from Tomorrow.io API."""
        if reset_num_api_requests:
            self._num_api_requests = 0

        if not additional_params:
            additional_params = {}

        ret_data = {}
        for i in range(0, len(fields), MAX_FIELDS_PER_REQUEST):
            data = await self._call_api(
                {
                    "timesteps": [TIMESTEP_CURRENT],
                    "fields": fields[i : i + MAX_FIELDS_PER_REQUEST],
                    **additional_params,
                }
            )
            try:
                ret_data.update(data["data"]["timelines"][0]["intervals"][0]["values"])
            except LookupError as error:
                raise UnknownException(data) from error

        return ret_data

    async def forecast(
        self,
        timesteps: List[timedelta],
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        reset_num_api_requests: bool = True,
        **additional_params,
    ) -> Dict[str, List[Dict[str, Any]]]:
        """Return forecast data from Tomorrow.io's API for a given time period."""
        if reset_num_api_requests:
            self._num_api_requests = 0

        params: Dict[str, Any] = {
            "timesteps": [_timedelta_to_str(timestep) for timestep in timesteps],
            **additional_params,
        }

        if start_time:
            if not start_time.tzinfo:
                start_time.replace(tzinfo=timezone.utc)
        else:
            start_time = datetime.now(tz=timezone.utc)
        params["startTime"] = start_time.replace(microsecond=0).isoformat()

        if duration:
            params["endTime"] = (
                (start_time + duration).replace(microsecond=0).isoformat()
            )

        forecasts: Dict[str, List[Dict[str, Any]]] = {}
        for i in range(0, len(fields), MAX_FIELDS_PER_REQUEST):
            data = await self._call_api(
                {**params, "fields": fields[i : i + MAX_FIELDS_PER_REQUEST]}
            )
            try:
                for timeline in data["data"]["timelines"]:
                    forecast_type = _timestep_to_key(timeline["timestep"])
                    if forecast_type not in forecasts:
                        forecasts[forecast_type] = timeline["intervals"]
                        continue
                    for idx in range(0, len(forecasts[forecast_type])):
                        forecasts[forecast_type][idx]["values"].update(
                            timeline["intervals"][idx]["values"]
                        )
            except LookupError as error:
                raise UnknownException(data) from error

        return forecasts

    async def forecast_nowcast(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        timestep: int = 5,
        reset_num_api_requests: bool = True,
        **additional_params,
    ) -> List[Dict[str, Any]]:
        """Return forecast data from Tomorrow.io's NowCast API for a given time period."""
        forecasts = await TomorrowioV4.forecast(
            self,
            [timedelta(minutes=timestep)],
            fields,
            start_time=start_time,
            duration=duration,
            reset_num_api_requests=reset_num_api_requests,
            **additional_params,
        )
        return forecasts[NOWCAST]

    async def forecast_daily(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        reset_num_api_requests: bool = True,
        **additional_params,
    ) -> List[Dict[str, Any]]:
        """Return daily forecast data from Tomorrow.io's API for a given time period."""
        forecasts = await TomorrowioV4.forecast(
            self,
            [ONE_DAY],
            fields,
            start_time=start_time,
            duration=duration,
            reset_num_api_requests=reset_num_api_requests,
            **additional_params,
        )
        return forecasts[DAILY]

    async def forecast_hourly(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        reset_num_api_requests: bool = True,
        **additional_params,
    ) -> List[Dict[str, Any]]:
        """Return hourly forecast data from Tomorrow.io's API for a given time period."""
        forecasts = await TomorrowioV4.forecast(
            self,
            [ONE_HOUR],
            fields,
            start_time=start_time,
            duration=duration,
            reset_num_api_requests=reset_num_api_requests,
            **additional_params,
        )
        return forecasts[HOURLY]

    async def all_forecasts(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        nowcast_timestep: int = 5,
        reset_num_api_requests: bool = True,
        **additional_params,
    ) -> Dict[str, List[Dict[str, Any]]]:
        """Return all forecasts."""
        return await TomorrowioV4.forecast(
            self,
            [
                timedelta(minutes=nowcast_timestep),
                ONE_HOUR,
                ONE_DAY,
            ],
            fields,
            start_time=start_time,
            duration=duration,
            reset_num_api_requests=reset_num_api_requests,
            **additional_params,
        )

    async def realtime_and_all_forecasts(
        self,
        realtime_fields: List[str],
        all_forecasts_fields: Optional[List[str]] = None,
        nowcast_fields: Optional[List[str]] = None,
        hourly_fields: Optional[List[str]] = None,
        daily_fields: Optional[List[str]] = None,
        nowcast_timestep: int = 5,
        **additional_params,
    ) -> Dict[str, Dict[str, Any]]:
        """
        Return realtime weather and all forecasts.

        To get the same fields for all forecasts, use all_forecasts_fields. To get
        specific fields for specific forecast types, use the corresponding fields list.

        additional keyword arguments will be added as additional parameters, overriding
        existing parameters when applicable.
        """
        self._num_api_requests = 0
        if not (
            all_forecasts_fields or nowcast_fields or hourly_fields or daily_fields
        ):
            raise ValueError("At least one field list must be specified")
        if all_forecasts_fields and any(
            fields for fields in (nowcast_fields, hourly_fields, daily_fields)
        ):
            raise ValueError(
                "Either only all_forecasts_fields list must be specified or at least "
                "one of the other field lists"
            )

        forecasts: Dict[str, List[Dict[str, Any]]] = {}
        if all_forecasts_fields:
            forecasts = await TomorrowioV4.all_forecasts(
                self,
                all_forecasts_fields,
                nowcast_timestep=nowcast_timestep,
                reset_num_api_requests=False,
                **additional_params,
            )
        else:
            for fields, timestep in [
                (nowcast_fields, timedelta(minutes=nowcast_timestep)),
                (hourly_fields, ONE_HOUR),
                (daily_fields, ONE_DAY),
            ]:
                if fields:
                    forecasts.update(
                        await TomorrowioV4.forecast(
                            self,
                            [timestep],
                            fields,
                            reset_num_api_requests=False,
                            **additional_params,
                        )
                    )

        current = await TomorrowioV4.realtime(
            self, realtime_fields, reset_num_api_requests=False, **additional_params
        )
        return {CURRENT: current, FORECASTS: forecasts}


class TomorrowioV4Sync(TomorrowioV4):
    """Synchronous class to query the Tomorrow.io API."""

    def __init__(
        self,
        apikey: str,
        latitude: Union[int, float, str],
        longitude: Union[int, float, str],
        unit_system: str = "imperial",
    ) -> None:
        """Initialize Synchronous Tomorrow.io v4 API object."""
        super().__init__(apikey, latitude, longitude, unit_system)

    @async_to_sync
    async def realtime(
        self,
        fields: List[str],
        reset_num_api_requests: bool = True,
        **additional_params,
    ) -> Dict[str, Any]:
        """Return realtime weather conditions from Tomorrow.io API."""
        return await super().realtime(
            fields, reset_num_api_requests=reset_num_api_requests, **additional_params
        )

    @async_to_sync
    async def forecast(
        self,
        timesteps: List[timedelta],
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        reset_num_api_requests: bool = True,
        **additional_params,
    ) -> Dict[str, List[Dict[str, Any]]]:
        """Return forecast data from Tomorrow.io's API for a given time period."""
        return await super().forecast(
            timesteps,
            fields,
            start_time=start_time,
            duration=duration,
            reset_num_api_requests=reset_num_api_requests,
            **additional_params,
        )

    @async_to_sync
    async def forecast_nowcast(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        timestep: int = 5,
        reset_num_api_requests: bool = True,
        **additional_params,
    ) -> List[Dict[str, Any]]:
        """Return forecast data from Tomorrow.io's NowCast API for a given time period."""
        return await super().forecast_nowcast(
            fields,
            start_time=start_time,
            duration=duration,
            timestep=timestep,
            reset_num_api_requests=reset_num_api_requests,
            **additional_params,
        )

    @async_to_sync
    async def forecast_daily(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        reset_num_api_requests: bool = True,
        **additional_params,
    ) -> List[Dict[str, Any]]:
        """Return daily forecast data from Tomorrow.io's API for a given time period."""
        return await super().forecast_daily(
            fields,
            start_time=start_time,
            duration=duration,
            reset_num_api_requests=reset_num_api_requests,
            **additional_params,
        )

    @async_to_sync
    async def forecast_hourly(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        reset_num_api_requests: bool = True,
        **additional_params,
    ) -> List[Dict[str, Any]]:
        """Return hourly forecast data from Tomorrow.io's API for a given time period."""
        return await super().forecast_hourly(
            fields,
            start_time=start_time,
            duration=duration,
            reset_num_api_requests=reset_num_api_requests,
            **additional_params,
        )

    @async_to_sync
    async def all_forecasts(
        self,
        fields: List[str],
        start_time: Optional[datetime] = None,
        duration: Optional[timedelta] = None,
        nowcast_timestep: int = 5,
        reset_num_api_requests: bool = True,
        **additional_params,
    ) -> Dict[str, List[Dict[str, Any]]]:
        """Return all forecasts."""
        return await super().all_forecasts(
            fields,
            start_time=start_time,
            duration=duration,
            nowcast_timestep=nowcast_timestep,
            reset_num_api_requests=reset_num_api_requests,
            **additional_params,
        )

    @async_to_sync
    async def realtime_and_all_forecasts(
        self,
        realtime_fields: List[str],
        all_forecasts_fields: Optional[List[str]] = None,
        nowcast_fields: Optional[List[str]] = None,
        hourly_fields: Optional[List[str]] = None,
        daily_fields: Optional[List[str]] = None,
        nowcast_timestep: int = 5,
        **additional_params,
    ) -> Dict[str, Dict[str, Any]]:
        """
        Return realtime weather and all forecasts.

        To get the same fields for all forecasts, use all_forecasts_fields. To get
        specific fields for specific forecast types, use the corresponding fields list.
        """
        return await super().realtime_and_all_forecasts(
            realtime_fields=realtime_fields,
            all_forecasts_fields=all_forecasts_fields,
            nowcast_fields=nowcast_fields,
            hourly_fields=hourly_fields,
            daily_fields=daily_fields,
            nowcast_timestep=nowcast_timestep,
            **additional_params,
        )



"""pytomorrowio helper functions."""

import asyncio
from functools import wraps


def async_to_sync(f):
    """Decorator to run async function as sync."""

    @wraps(f)
    def wrapper(*args, **kwargs):
        return asyncio.run(f(*args, **kwargs))

    return wrapper



"""Test helpers"""
import json
from http import HTTPStatus
from types import SimpleNamespace
from typing import List, Mapping, Optional, Union

from aiohttp import (
    ClientSession,
    TraceConfig,
    TraceRequestChunkSentParams,
    TraceRequestEndParams,
    web,
)

from .const import TEST_V4_PATH


async def create_session(
    aiohttp_client,
    file_name: Union[str, List[str]],
    *,
    headers: Optional[Mapping[str, str]] = None,
    status: HTTPStatus = HTTPStatus.OK,
):
    """Create aiohttp session that returns results from a file"""
    ns = SimpleNamespace(
        file_names=[file_name] if isinstance(file_name, str) else file_name,
        index=0,
        headers=headers,
        status=status,
    )

    async def response_handler(request):  # pylint: disable=unused-argument
        if ns.index >= len(ns.file_names):
            return None  # it's too late in the pipe to generate connection exception

        file_name = ns.file_names[ns.index]
        ns.index += 1

        with open(f"tests/fixtures/{file_name}", "r", encoding="utf8") as file:
            text = file.read()

        return web.json_response(text=text, headers=ns.headers, status=ns.status)

    app = web.Application()
    app.router.add_post(TEST_V4_PATH, response_handler)
    return await aiohttp_client(app)


def create_trace_config() -> TraceConfig:
    """Create 'TraceConfig' for tracing aiohttp responses"""

    async def on_request_chunk_sent(
        session: ClientSession,  # pylint: disable=unused-argument
        trace_config_ctx: SimpleNamespace,
        params: TraceRequestChunkSentParams,
    ):
        trace_config_ctx.request_body = params.chunk

    async def on_request_end(
        session: ClientSession,  # pylint: disable=unused-argument
        trace_config_ctx: SimpleNamespace,
        params: TraceRequestEndParams,
    ):
        print("Request:")
        print(params.url)
        for k, v in params.headers.items():
            print(f"  {k}: {v}")
        if trace_config_ctx.request_body:
            print(trace_config_ctx.request_body)
        print()
        print("Response:")
        for k, v in sorted(params.response.headers.items()):
            print(f"  {k}: {v}")
        resp = await params.response.json()
        print(json.dumps(resp, indent=2))
        print()

    trace_config = TraceConfig()
    trace_config.on_request_chunk_sent.append(on_request_chunk_sent)
    trace_config.on_request_end.append(on_request_end)
    return trace_config
